'''Class to provide namespace manipulation.
'''

import json
import time
from eliot import start_action
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild


class LSSTNamespaceManager(LoggableChild):
    '''Class to provide namespace manipulation.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.namespace = None
        self.service_account = None  # Account for pod to run as

    def set_namespace(self, namespace):
        with start_action(action_type="set_namespace"):
            self.namespace = namespace

    def ensure_namespace(self):
        '''Here we make sure that the namespace exists, creating it if
        it does not.  That requires a ClusterRole that can list and create
        namespaces.

        If we create the namespace, we also create (if needed) a ServiceAccount
        within it to allow the user pod to spawn dask and workflow pods.

        '''
        with start_action(action_type="ensure_namespace"):
            namespace = self.namespace
            if not namespace or namespace == "default":
                raise ValueError("Will not use default namespace!")
            api = self.parent.api
            cfg = self.parent.config
            acd = 'argocd.argoproj.io/'
            ns = client.V1Namespace(
                metadata=client.V1ObjectMeta(
                    name=namespace,
                    labels={acd + 'instance': 'nublado-users'},
                    annotations={
                        acd + 'compare-options': 'IgnoreExtraneous',
                        acd + 'sync-options': 'Prune=false',
                    }))
            try:
                self.log.info(
                    "Attempting to create namespace '%s'" % namespace)
                api.create_namespace(ns)
            except ApiException as e:
                if e.status != 409:
                    estr = "Create namespace '%s' failed: %s" % (ns, str(e))
                    self.log.exception(estr)
                    raise
                else:
                    self.log.info("Namespace '%s' already exists." % namespace)
            # Wait for the namespace to actually appear before creating objects
            #  in it.
            self.wait_for_namespace()
            if cfg.allow_dask_spawn:
                self.log.debug("Ensuring namespaced service account.")
                self.ensure_namespaced_service_account()
            if self.parent.spawner.enable_namespace_quotas:
                # By the time we need this, quota will have been set, because
                #  we needed it for options form generation.
                self.log.debug("Determining resource quota.")
                qm = self.parent.quota_mgr
                qm.ensure_namespaced_resource_quota(qm.quota)
            self.log.debug("Namespace resources ensured.")

    def def_namespaced_account_objects(self):
        '''Define K8s objects for things we need in the namespace.
        '''
        with start_action(action_type="define_namespaced_account_objects"):
            namespace = self.namespace
            username = self.parent.user.escaped_name
            account = "{}-svcacct".format(username)
            self.service_account = account
            acd = 'argocd.argoproj.io/'
            md = client.V1ObjectMeta(
                name=account,
                labels={acd + 'instance': 'nublado-users'},
                annotations={
                    acd + 'compare-options': 'IgnoreExtraneous',
                    acd + 'sync-options': 'Prune=false',
                }
            )
            svcacct = client.V1ServiceAccount(metadata=md)
            # These rules let us manipulate Dask pods, Argo Workflows, and
            #  Multus CNI interfaces
            rules = [
                client.V1PolicyRule(
                    api_groups=["argoproj.io"],
                    resources=["workflows", "workflows/finalizers"],
                    verbs=["get", "list", "watch", "update", "patch", "create",
                           "delete"]
                ),
                client.V1PolicyRule(
                    api_groups=["argoproj.io"],
                    resources=["workflowtemplates",
                               "workflowtemplates/finalizers"],
                    verbs=["get", "list", "watch"],
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["secrets"],
                    verbs=["get"]
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["pods", "pods/exec", "services", "configmaps"],
                    verbs=["get", "list", "watch", "create", "delete",
                           "update", "patch"]
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["pods/log", "serviceaccounts"],
                    verbs=["get", "list", "watch"]
                ),
            ]
            role = client.V1Role(
                rules=rules,
                metadata=md)
            rbstr = 'rbac.authorization.k8s.io'
            rolebinding = client.V1RoleBinding(
                metadata=md,
                role_ref=client.V1RoleRef(api_group=rbstr,
                                          kind="Role",
                                          name=account),
                subjects=[client.V1Subject(
                    kind="ServiceAccount",
                    name=account,
                    namespace=namespace)]
            )
            return svcacct, role, rolebinding

    def ensure_namespaced_service_account(self):
        '''Create a service account with role and rolebinding to allow it
        to manipulate resources in the namespace.
        '''
        with start_action(action_type="ensure_namespaced_service_account"):
            self.log.info("Ensuring namespaced service account.")
            namespace = self.namespace
            api = self.parent.api
            rbac_api = self.parent.rbac_api
            svcacct, role, rolebinding = self.def_namespaced_account_objects()
            account = self.service_account
            try:
                self.log.info("Attempting to create service account.")
                api.create_namespaced_service_account(
                    namespace=namespace,
                    body=svcacct)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception("Create service account '{}' " +
                                       "in namespace '{}' " +
                                       "failed: '{}".format(account,
                                                            namespace, e))
                    raise
                else:
                    self.log.info("Service account '{}' " +
                                  "in namespace '{}' " +
                                  "already exists.".format(account, namespace))
            try:
                self.log.info("Attempting to create role in namespace.")
                rbac_api.create_namespaced_role(
                    namespace,
                    role)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception("Create role '%s' " % account +
                                       "in namespace '%s' " % namespace +
                                       "failed: %s" % str(e))
                    raise
                else:
                    self.log.info("Role '{}' already exists in " +
                                  "namespace '{}'.".format(account, namespace))
            try:
                self.log.info("Attempting to create rolebinding in namespace.")
                rbac_api.create_namespaced_role_binding(
                    namespace,
                    rolebinding)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception("Create rolebinding '%s'" % account +
                                       "in namespace '%s' " % namespace +
                                       "failed: %s", str(e))
                    raise
                else:
                    self.log.info("Rolebinding '%s' " % account +
                                  "already exists in '%s'." % namespace)

    def wait_for_namespace(self, timeout=30):
        '''Wait for namespace to be created.'''
        with start_action(action_type="wait_for_namespace"):
            namespace = self.namespace
            for dl in range(timeout):
                self.log.debug("Checking for namespace " +
                               "{} [{}/{}]".format(namespace, dl, timeout))
                nl = self.parent.api.list_namespace(timeout_seconds=1)
                for ns in nl.items:
                    nsname = ns.metadata.name
                    if nsname == namespace:
                        self.log.debug("Namespace {} found.".format(namespace))
                        return
                    self.log.debug(
                        "Namespace {} not present yet.".format(namespace))
                time.sleep(1)
            raise RuntimeError(
                "Namespace '{}' not created in {} seconds!".format(namespace,
                                                                   timeout))

    def maybe_delete_namespace(self):
        '''Here we try to delete the namespace.  If it has no non-dask
        running pods, and it's not the default namespace, we can delete it."

        This requires a cluster role that can delete namespaces.'''
        with start_action(action_type="maybe_delete_namespace"):
            self.log.debug("Attempting to delete namespace.")
            namespace = self.namespace
            if not namespace or namespace == "default":
                raise RuntimeError("Cannot delete default namespace!")
            podlist = self.parent.api.list_namespaced_pod(namespace)
            clear_to_delete = True
            if podlist and podlist.items:
                clear_to_delete = self._check_pods(podlist.items)
            if not clear_to_delete:
                self.log.info("Not deleting namespace '%s'" % namespace)
                return False
            self.log.info("Deleting namespace '%s'" % namespace)
            self.parent.api.delete_namespace(namespace)
            return True

    def _check_pods(self, items):
        with start_action(action_type="_check_pods"):
            namespace = self.namespace
            for i in items:
                if i and i.status:
                    phase = i.status.phase
                    if (phase == "Running" or phase == "Unknown"
                            or phase == "Pending"):
                        pname = i.metadata.name
                        if pname.startswith("dask-"):
                            self.log.debug(
                                ("Abandoned dask pod '{}' can be " +
                                 "reaped.").format(pname))
                            # We can murder abandoned dask pods
                            continue
                        self.log.warning(("Pod in state '{}'; cannot delete " +
                                          "namespace '{}'.").format(phase,
                                                                    namespace))
                        return False
            # FIXME check on workflows as well.
            return True

    def destroy_namespaced_resource_quota(self):
        '''Remove quotas from namespace.
        You don't usually have to call this, since it will get
        cleaned up as part of namespace deletion.
        '''
        with start_action(action_type="destroy_namespaced_resource_quota"):
            namespace = self.get_user_namespace()
            api = self.parent.api
            qname = "quota-" + namespace
            dopts = client.V1DeleteOptions()
            self.log.info("Deleting resourcequota '%s'" % qname)
            api.delete_namespaced_resource_quota(qname, namespace, dopts)

    def delete_namespaced_svcacct_objs(self):
        '''Remove service accounts, roles, and rolebindings from namespace.
        You don't usually have to call this, since they will get
         cleaned up as part of namespace deletion.
        '''
        with start_action(action_type="delete_namespaced_svcacct_objs"):
            namespace = self.namespace
            account = self.service_account
            if not account:
                self.log.info("Service account not defined.")
                return
            dopts = client.V1DeleteOptions()
            self.log.info("Deleting service accounts/role/rolebinding " +
                          "for %s" % namespace)
            self.parent.rbac_api.delete_namespaced_role_binding(
                account,
                namespace,
                dopts)
            self.parent.rbac_api.delete_namespaced_role(
                account,
                namespace,
                dopts)
            self.parent.api.delete_namespaced_service_account(
                account,
                namespace,
                dopts)

    def dump(self):
        '''Return dict for pretty-printing.
        '''
        nd = {"namespace": self.namespace,
              "service_account": self.service_account,
              "parent": str(self.parent)}
        return nd

    def toJSON(self):
        return json.dumps(self.dump())



import json
from eliot import start_action
from ..utils import str_true, sanitize_dict
from .. import LoggableChild


class LSSTEnvironmentManager(LoggableChild):
    '''Class to hold LSST-specific, but not user-specific, environment.
    The user- specific pieces (from the volume manager and the quota
    manager) will be injected at spawn time.

    The environment should be regarded as immutable after creation.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.env = {}

    def create_pod_env(self):
        '''Return a dict mapping string to string for injection into the
        pod environment.
        '''
        with start_action(action_type="create_pod_env"):
            cfg = self.parent.config
            env = {}
            env['DEBUG'] = str_true(cfg.debug)
            cpug = float(cfg.tiny_cpu_max) / float(cfg.lab_size_range)
            env['CPU_GUARANTEE'] = str(cpug)
            env['MEM_GUARANTEE'] = "{}M".format(int(
                cpug * int(cfg.mb_per_cpu)))
            env['LAB_SIZE_RANGE'] = str(cfg.lab_size_range)
            env['CULL_TIMEOUT'] = cfg.cull_timeout
            env['CULL_POLICY'] = cfg.cull_policy
            env['RESTRICT_DASK_NODES'] = str_true(cfg.restrict_dask_nodes)
            env['LAB_NODEJS_MAX_MEM'] = cfg.lab_nodejs_max_mem
            env['NODE_OPTIONS'] = cfg.lab_node_options
            env['EXTERNAL_HUB_URL'] = cfg.external_hub_url
            env['HUB_ROUTE'] = cfg.hub_route
            env['EXTERNAL_HUB_URL'] = cfg.external_hub_url
            env['EXTERNAL_URL'] = cfg.external_hub_url
            env['EXTERNAL_INSTANCE_URL'] = cfg.external_instance_url
            env['FIREFLY_ROUTE'] = cfg.firefly_route
            env['JS9_ROUTE'] = cfg.js9_route
            env['API_ROUTE'] = cfg.api_route
            env['TAP_ROUTE'] = cfg.tap_route
            env['SODA_ROUTE'] = cfg.soda_route
            env['WORKFLOW_ROUTE'] = cfg.workflow_route
            env['EXTERNAL_FIREFLY_URL'] = cfg.external_firefly_url
            env['EXTERNAL_JS9_URL'] = cfg.external_js9_url
            env['EXTERNAL_API_URL'] = cfg.external_api_url
            env['EXTERNAL_TAP_URL'] = cfg.external_tap_url
            env['EXTERNAL_SODA_URL'] = cfg.external_soda_url
            env['EXTERNAL_WORKFLOW_URL'] = cfg.external_workflow_url
            env['AUTO_REPO_URLS'] = cfg.auto_repo_urls
            # Now clean up the env hash by removing any keys with empty values
            cleaned = self._clean_env(env)
            sanitized = self._sanitize(cleaned)
            self.log.debug("create_env yielded:\n.{}".format(sanitized))
            self.pod_env = cleaned

    def _mem_workaround(self, mem):
        '''We need to stop appending "M" to the dask template.  For now
        we return size-in-megabytes-with-no-suffix.
        '''
        with start_action(action_type="_mem_workaround"):
            if not mem:
                return "1"
            mem_s = str(mem)
            last_c = mem_s[-1].upper()
            # If we get ever-so-precious "*i", deal with it.  "M" also means
            #  2^20 here, not 10^6.  Deal with it.
            if last_c.isdigit():
                return mem_s
            mem_s = mem_s[:-1]
            if last_c == "i":
                mem_s = mem_s[:-1]
                last_c = mem_s[-1]
            mem_i = None
            try:
                mem_i = int(mem_s)
            except ValueError:
                return "1"
            if mem_i < 1:
                mem_i = 1
            if last_c == "K":
                mem_i = int(mem_i / 1024)
                if mem_i < 1:
                    mem_i = 1
            elif last_c == "G":
                mem_i = 1024 * mem_i
            elif last_c == "T":
                mem_i = 1024 * 1024 * mem_i
            else:
                # Assume M
                pass
            return str(mem_i)

    def _clean_env(self, env):
        with start_action(action_type="_clean_env"):
            return {str(k): str(v) for k, v in env.items() if
                    (v is not None and str(v) != '')}

    def get_env(self):
        '''Return the whole stored environment to caller as a dict.
        '''
        with start_action(action_type="get_env"):
            return self.pod_env

    def get_env_key(self, key):
        '''Return value of a specific key in the stored environment to caller.
        '''
        with start_action(action_type="get_env_key"):
            return self.pod_env.get(key)

    def _sanitize(self, incoming):
        with start_action(action_type="_sanitize"):
            sensitive = ['ACCESS_TOKEN', 'GITHUB_ACCESS_TOKEN',
                         'JUPYTERHUB_API_TOKEN', 'JPY_API_TOKEN']
            return sanitize_dict(incoming, sensitive)

    def dump(self):
        '''Return contents as a dict for pretty-printing.
        '''
        ed = {"parent": str(self.parent),
              "pod_env": self._sanitize(self.pod_env)
              }
        return ed

    def toJSON(self):
        return json.dumps(self.dump())



import json
from eliot import start_action
from .. import Loggable
from .apimanager import LSSTAPIManager
from .authmanager import LSSTAuthManager
from .envmanager import LSSTEnvironmentManager
from .namespacemanager import LSSTNamespaceManager
from .optionsformmanager import LSSTOptionsFormManager
from .quotamanager import LSSTQuotaManager
from .volumemanager import LSSTVolumeManager
from .workflowmanager import LSSTWorkflowManager


class LSSTMiddleManager(Loggable):
    '''The LSSTMiddleManager is a class that holds references to various
    LSST-specific management objects and delegates requests to them.
    The idea is that an LSST Spawner, or an LSST Workflow Manager,
    could instantiate a single LSSTMiddleManager, which would then be
    empowered to perform all LSST-specific operations, reducing
    configuration complexity.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = kwargs.pop('config', None)
        self.parent = kwargs.pop('parent', None)
        self.log.info(
            "Parent of LSST Middle Manager is '{}'".format(self.parent))
        self.authenticator = kwargs.pop('authenticator', None)
        self.api_mgr = LSSTAPIManager(parent=self)
        self.auth_mgr = LSSTAuthManager(parent=self)
        self.env_mgr = LSSTEnvironmentManager(parent=self)
        self.namespace_mgr = LSSTNamespaceManager(parent=self)
        self.optionsform_mgr = LSSTOptionsFormManager(parent=self)
        self.quota_mgr = LSSTQuotaManager(parent=self)
        self.volume_mgr = LSSTVolumeManager(parent=self)
        self.workflow_mgr = LSSTWorkflowManager(parent=self)
        self.api = self.api_mgr.api
        self.rbac_api = self.api_mgr.rbac_api
        self.wf_api = self.api_mgr.wf_api

    def ensure_resources(self):
        '''Delegate to namespace manager.
        '''
        with start_action(action_type="ensure_resources"):
            self.namespace_mgr.ensure_namespace()

    def dump(self):
        '''Return contents dict to pretty-print.
        '''
        md = {"parent": str(self.parent),
              "authenticator": str(self.authenticator),
              "spawner": str(self.spawner),
              "user": str(self.user),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              "wf_api": str(self.wf_api),
              "config": self.config.dump(),
              "api_mgr": self.api_mgr.dump(),
              "auth_mgr": self.auth_mgr.dump(),
              "env_mgr": self.env_mgr.dump(),
              "optionsform_mgr": self.optionsform_mgr.dump(),
              "quota_mgr": self.quota_mgr.dump(),
              "volume_mgr": self.volume_mgr.dump(),
              "workflow_mgr": self.workflow_mgr.dump()
              }
        return md

    def toJSON(self):
        return json.dumps(self.dump())



import json
import os
from eliot import start_action
from kubernetes.client import V1ResourceQuotaSpec
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild


class LSSTQuotaManager(LoggableChild):
    '''Quota support for LSST LSP Jupyterlab and Dask pods.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.quota = {}
        self.custom_resources = {}
        self.resourcemap = None

    def define_resource_quota_spec(self):
        '''We're going to return a resource quota spec that checks whether we
        have a custom resource map and uses that information.  If we do not
        then we use the quota from our parent's config object.

        Note that you could get a lot fancier, and check the user group
        memberships to determine what class a user belonged to, or some other
        more-sophisticated-than-one-size-fits-all quota mechanism.
        '''
        with start_action(action_type="define_resource_quota_spec"):
            self.log.debug("Calculating default resource quotas.")
            om = self.parent.optionsform_mgr
            sizemap = om.sizemap
            # sizemap is an ordered dict, and we want the last-inserted one,
            #  which is the biggest
            big = list(sizemap.keys())[-1]
            cpu = sizemap[big]['cpu']
            cfg = self.parent.config
            max_dask_workers = 0
            if cfg.allow_dask_spawn:
                max_dask_workers = cfg.max_dask_workers
            mem_per_cpu = cfg.mb_per_cpu
            total_cpu = (1 + max_dask_workers) * cpu
            total_mem = str(int(total_cpu * mem_per_cpu + 0.5)) + "Mi"
            total_cpu = str(int(total_cpu + 0.5))
            self.log.debug("Default quota sizes: CPU %r, mem %r" % (
                total_cpu, total_mem))
            self._set_custom_user_resources()
            if self.custom_resources:
                self.log.debug("Have custom resources.")
                cpuq = self.custom_resources.get("cpu_quota")
                if cpuq:
                    self.log.debug("Overriding CPU quota.")
                    total_cpu = str(cpuq)
                memq = self.custom_resources.get("mem_quota")
                if memq:
                    self.log.debug("Overriding memory quota.")
                    total_mem = str(memq) + "Mi"
            self.log.debug("Determined quota sizes: CPU %r, mem %r" % (
                total_cpu, total_mem))
            qs = V1ResourceQuotaSpec(
                hard={"limits.cpu": total_cpu,
                      "limits.memory": total_mem})
            self.quota = qs.hard

    def _set_custom_user_resources(self):
        '''Create custom resource definitions for user.
        '''
        with start_action(action_type="_set_custom_user_resources"):
            if not self.resourcemap:
                self.log.debug("No resource map found; generating.")
                self._create_resource_map()
            if not self.resourcemap:
                self.log.warning("No resource map; cannot generate quota.")
                return
            resources = {
                "size_index": 0,
                "cpu_quota": 0,
                "mem_quota": 0
            }
            gnames = self.parent.user.groups
            uname = self.parent.user.escaped_name
            for resdef in self.resourcemap:
                apply = False
                if resdef.get("disabled"):
                    continue
                candidate = resdef.get("resources")
                if not candidate:
                    continue
                self.log.debug(
                    "Considering candidate resource map {}".format(resdef))
                ruser = resdef.get("user")
                rgroup = resdef.get("group")
                if ruser and ruser == uname:
                    self.log.debug("User resource map match.")
                    apply = True
                if rgroup and rgroup in gnames:
                    self.log.debug("Group resource map match.")
                    apply = True
                if apply:
                    for fld in ["size_index", "cpu_quota", "mem_quota"]:
                        vv = candidate.get(fld)
                        if vv and vv > resources[fld]:
                            resources[fld] = vv
                        self.log.debug(
                            "Setting custom resources '{}'".format(resources) +
                            "for user '{}'".format(uname))
                        self.custom_resources = resources

    def _create_resource_map(self):
        with start_action(action_type="_create_resource_map"):
            resource_file = self.parent.config.resource_map
            if not os.path.exists(resource_file):
                nf_msg = ("Could not find resource definition file" +
                          " at '{}'".format(resource_file))
                self.log.error(nf_msg)
                return
            with open(resource_file, "r") as rf:
                resmap = json.load(rf)
            self.resourcemap = resmap

    # Brought in from namespacedkubespawner
    def ensure_namespaced_resource_quota(self, quotaspec):
        '''Create K8s quota object if necessary.
        '''
        with start_action(action_type="ensure_namespaced_resource_quota"):
            self.log.debug("Entering ensure_namespaced_resource_quota()")
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.api
            if namespace == "default":
                self.log.error("Will not create quota for default namespace!")
                return
            quota = client.V1ResourceQuota(
                metadata=client.V1ObjectMeta(
                    name="quota",
                ),
                spec=quotaspec
            )
            self.log.info("Creating quota: %r" % quota)
            try:
                api.create_namespaced_resource_quota(namespace, quota)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception("Create resourcequota '%s'" % quota +
                                       "in namespace '%s' " % namespace +
                                       "failed: %s", str(e))
                    raise
                else:
                    self.log.debug("Resourcequota '%r' " % quota +
                                   "already exists in '%s'." % namespace)

    def destroy_namespaced_resource_quota(self):
        '''Destroys the Kubernetes namespaced resource quota.
        You don't usually have to call this, since it will get
        cleaned up as part of namespace deletion.
        '''
        with start_action(action_type="destroy_namespaced_resource_quota"):
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.api
            qname = "quota-" + namespace
            dopts = client.V1DeleteOptions()
            self.log.info("Deleting resourcequota '%s'" % qname)
            api.delete_namespaced_resource_quota(
                qname, namespace, dopts)

    def dump(self):
        '''Return contents dict for pretty-printing/aggregation.
        '''
        qd = {"parent": str(self.parent),
              "quota": self.quota,
              "custom_resources": self.custom_resources,
              "resourcemap": self.resourcemap}
        return qd

    def toJSON(self):
        return json.dumps(self.dump())
